/*****************************************************************************\
|                                               ( )_  _                       |
|    _ _    _ __   _ _    __    ___ ___     _ _ | ,_)(_)  ___   ___     _     |
|   ( '_`\ ( '__)/'_` ) /'_ `\/' _ ` _ `\ /'_` )| |  | |/',__)/' v `\ /'_`\   |
|   | (_) )| |  ( (_| |( (_) || ( ) ( ) |( (_| || |_ | |\__, \| (˅) |( (_) )  |
|   | ,__/'(_)  `\__,_)`\__  |(_) (_) (_)`\__,_)`\__)(_)(____/(_) (_)`\___/'  |
|   | |                ( )_) |                                                |
|   (_)                 \___/'                                                |
|                                                                             |
| General Bots Copyright (c) Pragmatismo.io. All rights reserved.             |
| Licensed under the AGPL-3.0.                                                |
|                                                                             |
| According to our dual licensing model, this program can be used either      |
| under the terms of the GNU Affero General Public License, version 3,        |
| or under a proprietary license.                                             |
|                                                                             |
| The texts of the GNU Affero General Public License with an additional       |
| permission and of our proprietary license can be found at and               |
| in the LICENSE file you have received along with this program.              |
|                                                                             |
| This program is distributed in the hope that it will be useful,             |
| but WITHOUT ANY WARRANTY, without even the implied warranty of              |
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                |
| GNU Affero General Public License for more details.                         |
|                                                                             |
| "General Bots" is a registered trademark of Pragmatismo.io.                 |
| The licensing of the program under the AGPLv3 does not imply a              |
| trademark license. Therefore any rights, title and interest in              |
| our trademarks remain entirely with us.                                     |
|                                                                             |
\*****************************************************************************/
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
exports.DialogKeywords = void 0;
const botlib_1 = require("botlib");
const urlJoin = require("url-join");
const app_1 = require("../../../src/app");
const SecService_1 = require("../../security.gbapp/services/SecService");
const SystemKeywords_1 = require("./SystemKeywords");
/**
 * Base services of conversation to be called by BASIC which
 * requries step variable to work.
 */
class DialogKeywords {
    /**
     * When creating this keyword facade, a bot instance is
     * specified among the deployer service.
     */
    constructor(min, deployer) {
        this.min = min;
        this.internalSys = new SystemKeywords_1.SystemKeywords(min, deployer);
    }
    /**
     * Base reference of system keyword facade, called directly
     * by the script.
     */
    sys() {
        return this.internalSys;
    }
    /**
     * Returns the today data filled in dd/mm/yyyy or mm/dd/yyyy.
     *
     * @example x = TODAY
     */
    async getToday(step) {
        let d = new Date(), month = '' + (d.getMonth() + 1), day = '' + d.getDate(), year = d.getFullYear();
        if (month.length < 2) {
            month = '0' + month;
        }
        if (day.length < 2) {
            day = '0' + day;
        }
        const locale = step.context.activity.locale;
        switch (locale) {
            case 'pt-BR':
                return [day, month, year].join('/');
            case 'en-US':
                return [month, day, year].join('/');
            default:
                return [year, month, day].join('/');
        }
    }
    /**
     * Quits the dialog, currently required to get out of VM context.
     *
     * @example EXIT
     */
    async exit(step) {
        await step.endDialog();
    }
    /**
     * Returns current time in format hh:dd.
     *
     * @example SAVE "file.xlsx", name, email, NOW
     *
     */
    async getNow() {
        const nowUTC = new Date();
        const now = new Date((typeof nowUTC === 'string' ?
            new Date(nowUTC) :
            nowUTC).toLocaleString('en-US', { timeZone: process.env.DEFAULT_TIMEZONE }));
        return now.getHours() + ':' + now.getMinutes();
    }
    /**
     * Sends a file to a given mobile.
     *
     * @example SEND FILE TO "+199988887777", "image.jpg", caption
     *
     */
    async sendFileTo(step, mobile, filename, caption) {
        botlib_1.GBLog.info(`BASIC: SEND FILE TO '${mobile}', filename '${filename}'.`);
        return await this.internalSendFile(null, mobile, filename, caption);
    }
    /**
     * Sends a file to the current user.
     *
     * @example SEND FILE "image.jpg"
     *
     */
    async sendFile(step, filename, caption) {
        return await this.internalSendFile(step, null, filename, caption);
    }
    /**
     * Defines the current language of the bot conversation.
     *
     * @example SET LANGUAGE "pt"
     *
     */
    async setLanguage(step, language) {
        const user = await this.min.userProfile.get(step.context, {});
        const sec = new SecService_1.SecService();
        user.systemUser = await sec.updateUserLocale(user.systemUser.userId, language);
        await this.min.userProfile.set(step.context, user);
    }
    /**
     * Returns the name of the user acquired by WhatsApp API.
     */
    async userName(step) {
        return step.context.activity.from.name;
    }
    /**
     * OBSOLETE.
     */
    async getFrom(step) {
        return await this.userMobile(step);
    }
    /**
     * Returns current mobile number from user in conversation.
     *
     * @example SAVE "file.xlsx", name, email, MOBILE
     *
     */
    async userMobile(step) {
        if (isNaN(step.context.activity.from.id)) {
            return 'No mobile available.';
        }
        else {
            return step.context.activity.from.id;
        }
    }
    /**
     * Shows the subject menu to the user
     *
     * @example MENU
     *
     */
    async showMenu(step) {
        return await step.beginDialog('/menu');
    }
    /**
     * Performs the transfer of the conversation to a human agent.
     *
     * @example TRANSFER
     *
     */
    async transfer(step) {
        return await step.beginDialog('/t');
    }
    /**
     * Hears something from user and put it in a variable
     *
     * @example HEAR name
     *
     */
    async hear(step, promise, previousResolve, kind, ...args) {
        function random(low, high) {
            return Math.random() * (high - low) + low;
        }
        const idPromise = random(0, 120000000);
        this.min.cbMap[idPromise] = {};
        this.min.cbMap[idPromise].promise = promise;
        const opts = { id: idPromise, previousResolve: previousResolve, kind: kind, args };
        if (previousResolve !== undefined) {
            previousResolve(opts);
        }
        else {
            await step.beginDialog('/hear', opts);
        }
    }
    /**
     * Talks to the user by using the specified text.
     */
    async talk(step, text) {
        return await this.min.conversationalService.sendText(this.min, step, text);
    }
    /**
     * Processes the sending of the file.
     */
    async internalSendFile(step, mobile, filename, caption) {
        if (filename.indexOf('.md') > -1) {
            botlib_1.GBLog.info(`BASIC: Sending the contents of ${filename} markdown to mobile ${mobile}.`);
            const md = await this.min.kbService.getAnswerTextByMediaName(this.min.instance.instanceId, filename);
            await this.min.conversationalService.sendMarkdownToMobile(this.min, step, mobile, md);
        }
        else {
            botlib_1.GBLog.info(`BASIC: Sending the file ${filename} to mobile ${mobile}.`);
            const url = urlJoin(app_1.GBServer.globals.publicAddress, 'kb', `${this.min.botId}.gbai`, `${this.min.botId}.gbkb`, 'assets', filename);
            await this.min.conversationalService.sendFile(this.min, step, mobile, url, caption);
        }
    }
}
exports.DialogKeywords = DialogKeywords;
//# sourceMappingURL=DialogKeywords.js.map